# This doesn't change much but occassionally
# check these values to ensure they reflect current prices.
PRICES = {
    # Price per 1 million characters based on the TTS engine used.
    'standard': 4.00,
    'neural': 16.00,
    'generative': 30.00,
    'long-form': 100.00,
}
